##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Base classes for unit tests
"""
import unittest

import MySQLdb

from .dummy import FakeConnection

DB = 'zmysqldatest'
DB_USER = 'zmysqldatest'
DB_PASSWORD = 'zmysqldatest'
DB_CONN_STRING = '%s %s %s' % (DB, DB_USER, DB_PASSWORD)

TABLE_NAME = 'test_zmysqlda'
TABLE_COL_INT = 'c_int'
TABLE_COL_VARCHAR = 'c_varchar'

COL_INT = 'INT(10) UNSIGNED ZEROFILL NOT NULL DEFAULT 0 PRIMARY KEY'
COL_VARCHAR = 'VARCHAR(20)'

NO_MYSQL_MSG = 'Please see the documentation for running functional tests.'


def fake_connect(**kw):
    return FakeConnection(**kw)


def real_connect():
    return MySQLdb.connect(user=DB_USER, passwd=DB_PASSWORD, db=DB)


def _mySQLNotAvailable():
    """ Helper method for unittest skipping
    """
    try:
        connection = real_connect()
        connection.close()
        return False
    except MySQLdb.OperationalError as exc:  # noqa
        # print('Cannot connect to %s as %s: %s' % (DB, DB_USER, exc.args))
        return True


class PatchedConnectionTestsBase(unittest.TestCase):
    """ Tests that require faking out MySQLdb.connect
    """

    def setUp(self):
        super(PatchedConnectionTestsBase, self).setUp()

        from Products.ZMySQLDA.db import MySQLdb
        self.old_connect = MySQLdb.connect
        MySQLdb.connect = fake_connect

    def tearDown(self):
        super(PatchedConnectionTestsBase, self).tearDown()

        if getattr(self, 'conn', None):
            self.conn._v_database_connection = None
            self.conn._v_connected = None

        from Products.ZMySQLDA.db import MySQLdb
        MySQLdb.connect = self.old_connect
        from Products.ZMySQLDA.DA import database_connection_pool
        database_connection_pool.clear()


class MySQLRequiredLayer:

    @classmethod
    def setUp(cls):
        # Called only once for all tests in this layer
        # Connect to the database and (re-)create a test table,
        # but only if the test database is available to prevent tracebacks.
        # Tests classes using this layer should be skipped if the test
        # database is unavailable: "@unittest.skipIf(__mySQLNotAvailable())"
        if not _mySQLNotAvailable():
            dbconn = real_connect()
            dbconn.query('DROP TABLE IF EXISTS %s' % TABLE_NAME)
            sql = 'CREATE TABLE %s (%s %s, %s %s) ENGINE MEMORY'
            dbconn.query(sql % (TABLE_NAME, TABLE_COL_INT, COL_INT,
                                TABLE_COL_VARCHAR, COL_VARCHAR))
            dbconn.close()

    @classmethod
    def testSetUp(cls):
        from Products.ZMySQLDA import DA
        # Clean out the test table before every test
        if not _mySQLNotAvailable():
            dbconn = real_connect()
            dbconn.query('DELETE FROM %s' % TABLE_NAME)
            dbconn.close()

        # Clear the connections cache
        DA.database_connection_pool.clear()



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import os
from setuptools import setup, find_packages


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


setup(
    name="Products.ZMySQLDA",
    version=read('version.txt').strip(),
    license="ZPL 2.1",
    author="HOFFMANN+LIEBENBERG in association with "
            "SNTL Publishing, Andy Dustman, John Eikenberry",
    author_email="zms@sntl-publishing.com",
    url="https://github.com/zms-publishing/Products.ZMySQLDA",
    description="MySQL Zope adapter.",
    long_description="MySQL Database Adapter for Zope.",
    classifiers=[
        "Development Status :: 6 - Mature",
        "Environment :: Web Environment",
        "Framework :: Zope :: 4",
        "License :: OSI Approved :: Zope Public License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Database",
        "Topic :: Database :: Front-Ends",
    ],
    packages=find_packages(),
    include_package_data=True,
    namespace_packages=["Products"],
    zip_safe=False,
    install_requires=[
        "setuptools >= 36.2",
        "six",
        "mysqlclient",
        "Products.ZSQLMethods",
    ]
)



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" The ZODB-based MySQL Database Connection object
"""
import six
from six.moves._thread import allocate_lock

from AccessControl.class_init import InitializeClass
from AccessControl.Permissions import change_database_methods
from AccessControl.Permissions import use_database_methods
from AccessControl.Permissions import view_management_screens
from AccessControl.SecurityInfo import ClassSecurityInfo
from AccessControl.SecurityInfo import ModuleSecurityInfo
from App.special_dtml import HTMLFile
from Persistence import Persistent
from Shared.DC.ZRDB.Connection import Connection as ConnectionBase

from .db import DB
from .db import DBPool
from .permissions import add_zmysql_database_connections
from .utils import TableBrowser
from .utils import table_icons

# Connection Pool for connections to MySQL.
# Maps one mysql client connection to one DA object instance.
database_connection_pool_lock = allocate_lock()
database_connection_pool = {}

# Combining connection pool with the DB pool gets you
# one DA/connection per connection with 1 DBPool with 1 DB/thread
# pool_id -> DA -> DBPool -> thread id -> DB
# dc_pool[pool_id] == DBPool_instance
# DBPool_instance[thread id] == DB instance


class Connection(ConnectionBase):
    """ Zope database adapter for MySQL/MariaDB
    """
    meta_type = "Z MySQL Database Connection"
    security = ClassSecurityInfo()
    zmi_icon = 'fas fa-database'

    auto_create_db = True
    use_unicode = False
    charset = None
    _v_connected = ""
    _isAnSQLConnection = 1
    info = None

    security.declareProtected(view_management_screens, 'manage_browse')
    manage_browse = HTMLFile("www/browse", globals())

    security.declareProtected(change_database_methods, 'manage_properties')
    manage_properties = HTMLFile("www/connectionEdit", globals())

    manage_options = ConnectionBase.manage_options + (
        {"label": "Browse", "action": "manage_browse"},)

    def __init__(self, id, title, connection_string, check, use_unicode=None,
                 charset=None, auto_create_db=None):
        """ Instance setup. Optionally opens the connection.

        :string: id -- The id of the ZMySQLDA Connection

        :string: title -- The title of the ZMySQLDA Connection

        :string: connection_string -- The connection string describes how to
                                      connect to the relational database.
                                      See the documentation for details.

        :bool: check -- Check if the database connection can be opened after
                        instantiation.

        :bool: use_unicode -- If set to ``True``, values from columns of type
                              ``CHAR``, ``VARCHAR`` and ``TEXT`` are returned
                              as unicode strings by the database backend.
                              Combined with the hardcoded ``utf8`` character
                              set of this package the setting allows you to
                              control the character set of database return
                              values better. Default: False.

        :string: charset -- The character set for the connection. MySQL/MariaDB
                            will encode query results to this character set.

                            On Python 2, both utf8 and Latin1 will work. On
                            Python 3, only utf8 will work.

                            Default on Python 2: Latin1 when ``use_unicode``
                            is off, utf8 otherwise
                            Default on Python 3: utf8

        :bool: auto_create_db -- If the database given in ``connection_string``
                                 does not exist, create it automatically.
                                 Default: False.
        """
        self.use_unicode = bool(use_unicode)
        self.charset = charset
        self.auto_create_db = bool(auto_create_db)
        return super(Connection, self).__init__(id, title, connection_string,
                                                check)

    def __setstate__(self, state):
        """ Skip super's __setstate__ as it connects which we don't want
            due to pool_key depending on acquisition.
        """
        Persistent.__setstate__(self, state)

    security.declareProtected(use_database_methods, 'factory')

    def factory(self):
        """ Base API. Returns factory method for DB connections.
        """
        return DB

    def _pool_key(self):
        """ Return key used for DA pool.
        """
        return self.getPhysicalPath()

    def _getConnection(self):
        """ Helper method to retrieve an existing or create a new connection
        """
        try:
            return self._v_database_connection
        except AttributeError:
            self.connect(self.connection_string)
            return self._v_database_connection

    security.declareProtected(use_database_methods, 'connect')

    def connect(self, conn_string):
        """ Base API. Opens connection to mysql. Raises if problems.

        :string: conn_string -- The database connection string
        """
        pool_key = self._pool_key()
        conn = database_connection_pool.get(pool_key)

        if conn is not None and conn.connection == conn_string:
            self._v_database_connection = conn
            self._v_connected = conn.connected_timestamp
        else:
            if conn is not None:
                conn.closeConnection()

            conn_pool = DBPool(self.factory(), create_db=self.auto_create_db,
                               use_unicode=self.use_unicode,
                               charset=self.charset)
            database_connection_pool_lock.acquire()
            try:
                conn = conn_pool(conn_string)
                database_connection_pool[pool_key] = conn
            finally:
                database_connection_pool_lock.release()

            self._v_database_connection = conn
            # XXX If date is used as such, it can be wrong because an
            # existing connection may be reused. But this is suposedly
            # only used as a marker to know if connection was successfull.
            self._v_connected = conn.connected_timestamp

        return self  # ??? why doesn't this return the connection ???

    security.declareProtected(use_database_methods, 'sql_quote__')

    def sql_quote__(self, sql_str, escapes={}):
        """ Base API. Used to massage SQL strings for use in queries.

        :string: sql_str -- The raw SQL string to transform.

        :dict: escapes -- Additional escape transformations.
                          Default: empty ``dict``.
        """
        connection = self._getConnection()

        if self.use_unicode and isinstance(sql_str, six.text_type):
            return connection.unicode_literal(sql_str)
        else:
            return connection.string_literal(sql_str)

    security.declareProtected(change_database_methods, 'manage_edit')

    def manage_edit(self, title, connection_string, check=None,
                    use_unicode=None, charset=None, auto_create_db=None):
        """ Edit the connection attributes through the Zope ZMI.

        :string: title -- The title of the ZMySQLDA Connection

        :string: connection_string -- The connection string describes how to
                                      connect to the relational database.
                                      See the documentation for details.

        :bool: check -- Check if the database connection can be opened after
                        instantiation. Default: False.

        :bool: use_unicode -- Use unicode internally. Default: False.

        :string: charset -- The character set for the connection. MySQL/MariaDB
                            will encode query results to this character set.

                            On Python 2, both utf8 and Latin1 will work. On
                            Python 3, only utf8 will work.

                            Default on Python 2: Latin1 when ``use_unicode``
                            is off, utf8 otherwise
                            Default on Python 3: utf8

        :bool: auto_create_db -- If the database given in ``connection_string``
                                 does not exist, create it automatically.
                                 Default: False.
        """
        self.use_unicode = bool(use_unicode)
        self.charset = charset
        self.auto_create_db = bool(auto_create_db)

        return super(Connection, self).manage_edit(title, connection_string,
                                                   check=check)

    security.declareProtected(view_management_screens, 'tpValues')

    def tpValues(self):
        """ Support the DTML ``tree`` tag

        Used in the Zope ZMI ``Browse`` tab
        """
        t_list = []
        connection = self._getConnection()

        for t_info in connection.tables(rdb=0):
            try:
                t_browser = TableBrowser()
                t_browser.__name__ = t_info["table_name"]
                t_browser._d = t_info
                t_browser._c = connection
                t_browser.icon = table_icons.get(t_info["table_type"], "text")
                t_list.append(t_browser)
            except Exception:
                pass

        return t_list


InitializeClass(Connection)


mod_security = ModuleSecurityInfo('Products.ZMySQLDA.DA')
mod_security.declareProtected(add_zmysql_database_connections,
                              'manage_addZMySQLConnectionForm')
manage_addZMySQLConnectionForm = HTMLFile("www/connectionAdd", globals())

mod_security.declareProtected(add_zmysql_database_connections,
                              'manage_addZMySQLConnection')


def manage_addZMySQLConnection(self, id, title, connection_string, check=None,
                               use_unicode=None, auto_create_db=None,
                               charset=None, REQUEST=None):
    """Factory function to add a connection object from the Zope ZMI.

    :string: id -- The id of the ZMySQLDA Connection

    :string: title -- The title of the ZMySQLDA Connection

    :string: connection_string -- The connection string describes how to
                                  connect to the relational database.
                                  See the documentation for details.

    :bool: check -- Check if the database connection can be opened after
                    instantiation. Default: False.

    :bool: use_unicode -- If set to ``True``, values from columns of type
                          ``CHAR``, ``VARCHAR`` and ``TEXT`` are returned
                          as unicode strings by the database backend.
                          Combined with the hardcoded ``utf8`` character
                          set of this package the setting allows you to
                          control the character set of database return values
                          better. Default: False.

    :string: charset -- The character set for the connection. MySQL/MariaDB
                        will encode query results to this character set.

                        On Python 2, both utf8 and Latin1 will work. On
                        Python 3, only utf8 will work.

                        Default on Python 2: Latin1 when ``use_unicode``
                        is off, utf8 otherwise
                        Default on Python 3: utf8

    :bool: auto_create_db -- If the database given in ``connection_string``
                             does not exist, create it automatically.
                             Default: False.

    :object: REQUEST -- The currently active Zope request object.
                        Default: None.
    """
    self._setObject(id,
                    Connection(id, title, connection_string, check,
                               use_unicode=use_unicode, charset=charset,
                               auto_create_db=auto_create_db))

    if REQUEST is not None:
        return self.manage_main(self, REQUEST)


mod_security.apply(globals())
